import re
import os
from langchain.llms import OpenAI, GPT4All
from langchain.chat_models import ChatOpenAI
from langchain.chains import LLMChain
from langchain.prompts import PromptTemplate
from langchain.prompts.chat import (
    HumanMessagePromptTemplate,
    ChatPromptTemplate
)
import yaml


CHAT_MODELS=["gpt-4", "gpt-4-0314", "gpt-4-32k", "gpt-4-32k-0314", "gpt-3.5-turbo", "gpt-3.5-turbo-0301"]
COMPLETION_MODELS=["text-davinci-003", "text-davinci-002", "text-curie-001", "text-babbage-001", "text-ada-001"]

def _extract_default_output(o):
    return o.strip()

def _extract_gpt4all_output(output_str):
    answer = "Answer:"
    index = output_str.find(answer)
    if index != -1:
        return output_str[index + len(answer):].strip()
    else:
        return ""

def _create_llm_chain(model_name, template, input_variables, temperature, max_tokens):
    extractor = _extract_default_output
    if model_name in CHAT_MODELS:
        llm = ChatOpenAI(model_name=model_name, temperature=temperature, max_tokens=max_tokens)
        prompt_template = ChatPromptTemplate.from_messages([HumanMessagePromptTemplate.from_template(template=template)])
    elif model_name in COMPLETION_MODELS:
        llm = OpenAI(model_name=model_name, temperature=temperature, max_tokens=max_tokens)
        prompt_template = PromptTemplate(template=template, input_variables=input_variables)
    elif os.path.isabs(model_name) or model_name.startswith('./') or model_name.startswith('../'):
        llm = GPT4All(model=model_name, temp=temperature, n_predict=max_tokens)
        # we wrap the original template in a request/response format so
        # - there's a better chance gpt4all does the right thing
        # - we can parse the output after response
        input_template=f'Question: {template}\nAnswer:'
        prompt_template = PromptTemplate(template=input_template, input_variables=input_variables)
        # we need to remove the question/answer template
        extractor = _extract_gpt4all_output
    else:
        raise ValueError(f"Invalid model_name: {model_name}. Please use a valid OpenAI model or provide a relative/absolute path for a GPT4All model.")

    return LLMChain(llm=llm, prompt=prompt_template), extractor

def _compare_output(output, expected_output):
    if expected_output['type'] == 'regex':
        pattern = re.compile(expected_output['value'])
        return pattern.match(output) is not None, None
    elif expected_output['type'] == 'string':
        return output == expected_output['value'], None
    elif expected_output['type'] == 'yaml':
        try:
            output_yaml = yaml.safe_load(output)
            expected_yaml = yaml.safe_load(expected_output['value'])
            return output_yaml == expected_yaml, None
        except yaml.YAMLError:
            return False
    elif expected_output['type'] == 'prompt':
        model = expected_output['model']
        llm = OpenAI(model_name=model, temperature=0, max_tokens=1800)
        out_value = expected_output['value']

        params = { 'input_text': output, 'conditions': out_value['conditions'] }
        current_file_path = os.path.abspath(__file__)
        current_dir_path = os.path.dirname(current_file_path)

        with open(os.path.join(current_dir_path, 'templates/conditions_on_text.yml'), 'r') as file:
            template = file.read()

        if model in CHAT_MODELS:
            llm = ChatOpenAI(model_name=model, temperature=0, max_tokens=1800)
            prompt_template = ChatPromptTemplate.from_messages([HumanMessagePromptTemplate.from_template(template=template)])
        else:
            llm = OpenAI(model_name=model, temperature=0, max_tokens=1800)
            prompt_template =  PromptTemplate(template=template, input_variables=['input_text', 'conditions'])

        llm_chain, extractor = LLMChain(llm=llm, prompt=prompt_template, output_key='validation_result')

        validation_result = llm_chain.predict(**params)

        try:
            validation_result_yaml = yaml.safe_load(validation_result)
            overall_pass = validation_result_yaml.get('pass', False)
            return overall_pass, validation_result_yaml
        except yaml.YAMLError:
            return False
    else:
        return False

def test_models(prompt_data, tests_data):
    template = prompt_data['template']
    input_variables = prompt_data['input_variables']

    tests = tests_data['tests']
    model_names = tests_data['model_names']
    temperature = tests_data['temperature']
    max_tokens = tests_data['max_tokens']

    results = {}

    for model_name in model_names:
        total = 0
        passes = 0
        model_results = []

        llm_chain, extractor = _create_llm_chain(model_name, template, input_variables, temperature, max_tokens)

        for item in tests:
            variables = item['variables']
            expected_output = item['expected_output']

            result = extractor(llm_chain.predict(**variables))

            comparison_result, extra_data = _compare_output(result, expected_output)

            if comparison_result:
                passes += 1

            test_result = {
                'inputs': variables,
                'output': result,
                'comparison_result': comparison_result,
                'extra_data': extra_data
            }
            model_results.append(test_result)

            total += 1

        results[model_name] = (model_results, passes, total)

    return results



from setuptools import setup, find_packages

setup(
    name="promptest",
    version="0.1.5",
    packages=find_packages(),
    install_requires=[
        "langchain",
        "PyYAML",
        "pyllamacpp"
    ],
    entry_points={
        "console_scripts": [
            "promptest=promptest.main:main",
        ]
    },
)



import argparse
import os
from datetime import datetime
import yaml
from .tester import test_models

# from https://github.com/yaml/pyyaml/issues/240
def str_presenter(dumper, data):
    """configures yaml for dumping multiline strings
    Ref: https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data"""
    if data.count('\n') > 0:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)

yaml.add_representer(str, str_presenter)
yaml.representer.SafeRepresenter.add_representer(str, str_presenter)

def print_results(results):
    for model_name, (_, passes, total) in results.items():
        ratio = passes / total
        print(f"Model: {model_name}, Passes: {passes}, Total: {total}, Passes/Total Ratio: {ratio:.2f}")

def save_test_results(model_name, model_results, passes, total, temperature, max_tokens, template, template_name):
    date = datetime.now().strftime('%Y_%m_%d_%H_%M')
    output_directory = f'prompt_test_results/{date}-{template_name}'

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    test_data = {
        'model': model_name,
        'results': model_results,
        'totals': {
            'passes': passes,
            'total': total
        },
        'temperature': temperature,
        'max_tokens': max_tokens,
        'prompt_template': template
    }

    model_file_name = os.path.basename(model_name)

    filename = f"{model_file_name}_test_result.yml"
    file_path = os.path.join(output_directory, filename)

    with open(file_path, 'w') as outfile:
        yaml.dump(test_data, outfile, default_flow_style=None, allow_unicode=True, width=float('inf'), line_break='\n')

def parse_input_files(prompt_path, tests_path):
    with open(prompt_path, 'r') as file:
        prompt_data = yaml.safe_load(file)

    with open(tests_path, 'r') as file:
        tests_data = yaml.safe_load(file)

    return prompt_data, tests_data

def main():
    parser = argparse.ArgumentParser(description='Test LLM models with a given template and inputs.')
    parser.add_argument('--prompt', '-p', type=str, required=True, help='Path to the prompt template YAML file')
    parser.add_argument('--tests', '-t', type=str, required=True, help='Path to the tests YAML file')

    args = parser.parse_args()

    prompt_data, tests_data = parse_input_files(args.prompt, args.tests)
    results = test_models(prompt_data, tests_data)

    template_name = os.path.splitext(os.path.basename(args.prompt))[0]

    for model_name, (model_results, passes, total) in results.items():
        save_test_results(model_name, model_results, passes, total, tests_data['temperature'], tests_data['max_tokens'], prompt_data['template'], template_name)

    print_results(results)
